/*
 * Scratchpad MMU - Test program module 1
 * Copyright (C) Jack Whitham 2009
 * http://www.jwhitham.org.uk/c/smmu.html
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation
 * (version 2.1 of the License only).
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#ifdef HAVE_STDINT
#include <stdint.h>
#endif

#ifdef __MICROBLAZE__
#include "mb_interface.h"
#endif

#include "hwdriver.h"
#include "print.h"
#include "simif.h"

unsigned rng ( void ) ;

#ifndef HAVE_STDINT
#ifndef uint16_t
#define uint16_t unsigned short
#endif
#ifndef uint8_t
#define uint8_t unsigned char
#endif
#endif

//#define VERBOSE     1 
#define NUM_CYCLES      0x100000
#define NUM_SUB_CYCLES  0x1000
#define MAX_ERRORS      0x14
#define USE_NUM_ENTRIES ( NUM_RW_ENTRIES + NUM_RO_ENTRIES - 1 )

typedef struct Size_Record_struct {
    unsigned    index ;
    unsigned    size ;
    unsigned    ref ;
    unsigned    loaded ;
    unsigned    copy_loc ;
    data_t *    true_address ;
} Size_Record ;

static int Run_Test_W ( char * name , Size_Record * tower , int width ) ;
static int Run_Test ( char * name , Size_Record * tower ) ;

int Towering ( void )
{
    Size_Record     tower [ USE_NUM_ENTRIES ] ;
    const unsigned  BIG_BSP_SIZE = 64 ;
    const unsigned  SMALL_BSP_SIZE = 8 ;
    unsigned        size , offset , i , size_diff ;


    /* Test with binary partition shape (start big) */
    size = BIG_BSP_SIZE ;
    offset = 0 ;
    for ( i = 0 ; i < USE_NUM_ENTRIES ; i ++ )
    {
        assert ( size > 0 ) ;
        tower [ i ] . index = offset + 1 ;
        tower [ i ] . size = size ;
        offset += size ;
        assert ( offset <= BIG_BSP_SIZE ) ;
        if ( offset == BIG_BSP_SIZE )
        {
            offset = 0 ;
            size /= 2 ;
        }
    }
    if ( ! Run_Test ( "BSP" , tower ) ) return 0 ;

    /* And from the other direction (start small) */
    size = 1 ;
    offset = size ;
    for ( i = 0 ; i < USE_NUM_ENTRIES ; i ++ )
    {
        assert ( size > 0 ) ;
        tower [ i ] . index = offset + 1 ;
        tower [ i ] . size = size ;
        offset += size ;
        if ( offset >= SMALL_BSP_SIZE )
        {
            offset = 0 ;
            size *= 2 ;
        }
    }
    if ( ! Run_Test ( "RBSP" , tower ) ) return 0 ;

    for ( size_diff = 1 ; size_diff <= 3 ; size_diff ++ )
    {
        /* Test with pyramid shape */
        for ( i = 0 ; i < USE_NUM_ENTRIES ; i ++ )
        {
            tower [ i ] . index = 1 + ( i * size_diff ) ;
            tower [ i ] . size = ( USE_NUM_ENTRIES - i ) * size_diff * 2 ;
        }
        if ( ! Run_Test ( "Pyramid" , tower ) ) return 0 ;

        /* Test with triangle shape */
        for ( i = 0 ; i < USE_NUM_ENTRIES ; i ++ )
        {
            tower [ i ] . index = 1 + ( i * size_diff ) ;
            tower [ i ] . size = (( 1 + USE_NUM_ENTRIES ) * size_diff ) -
                    tower [ i ] . index ;
        }
        if ( ! Run_Test ( "Triangle" , tower ) ) return 0 ;

        /* Test with triangle shape (other way round) */
        for ( i = 0 ; i < USE_NUM_ENTRIES ; i ++ )
        {
            tower [ i ] . index = 1 ;
            tower [ i ] . size = (( i + 1 ) * size_diff ) ;
        }
        if ( ! Run_Test ( "Rev Triangle" , tower ) ) return 0 ;

        /* Test with Lemmings bridge shape (left to right) */
        for ( i = 0 ; i < USE_NUM_ENTRIES ; i ++ )
        {
            tower [ i ] . index = 1 + i ;
            tower [ i ] . size = size_diff ;
        }
        if ( ! Run_Test ( "Lemmings L to R" , tower ) ) return 0 ;

        /* Test with Lemmings bridge shape (right to left) */
        for ( i = 0 ; i < USE_NUM_ENTRIES ; i ++ )
        {
            tower [ i ] . index = 1 + USE_NUM_ENTRIES - i ;
            tower [ i ] . size = size_diff ;
        }
        if ( ! Run_Test ( "Lemmings R to L" , tower ) ) return 0 ;
    }
    return 1 ;
}


static int Run_Test_W ( char * name , Size_Record * tower , int width )
{
    data_t *        shadow ;
    data_t *        smmu ;
    uint16_t *      shadow2 ;
    uint16_t *      smmu2 ;
    uint8_t *       shadow4 ;
    uint8_t *       smmu4 ;
    unsigned        KEY1 = rng () ^ ( rng () << 16 ) ;
    unsigned        KEY2 = rng () ;
    unsigned        min_index = rng () % 64 ;
    unsigned        copy_loc = min_index ;
    unsigned        space , i , j , sc ;
    unsigned        max_index , loaded ;
    unsigned        num_ro_open = 0 ;
    unsigned        num_rw_open = 0 ;

    assert (( width == 1 ) || ( width == 2 ) || ( width == 4 )) ;
    printf ( "\rTest name: " ) ;
    print ( name ) ;
    max_index = 0 ;
    for ( i = 0 ; i < USE_NUM_ENTRIES ; i ++ )
    {
        tower [ i ] . index += min_index ;
        tower [ i ] . ref = 0 ;
        tower [ i ] . loaded = 0 ;
        tower [ i ] . copy_loc = copy_loc ;
        assert ( tower [ i ] . size != 0 ) ;
        assert ( tower [ i ] . index != 0 ) ;
        assert ( tower [ i ] . size < SP_SIZE ) ;

        copy_loc += tower [ i ] . size ;
        j = tower [ i ] . index + tower [ i ] . size ;
        if ( j > max_index )
            max_index = j ;
    }
    max_index += 2 ;
    if ( min_index > 2 ) {
        min_index -= 2 ;
    } else {
        min_index = 0;
    }

    space = copy_loc ;
    printf ( " size: %c", width + '0');
    printf ( " scratchpad space: %d", space ) ;
    printf ( " memory space: %d", max_index ) ;
    printf ( "\r\n" ) ;
    assert ( copy_loc <= SP_SIZE ) ;
    assert ( max_index > 0 ) ;

    shadow = alloca ( max_index * sizeof ( data_t ) ) ;
    smmu = alloca ( max_index * sizeof ( data_t ) ) ;
    shadow2 = (uint16_t *) shadow ;
    shadow4 = (uint8_t *) shadow ;
    smmu2 = (uint16_t *) smmu ;
    smmu4 = (uint8_t *) smmu ;


    for ( i = 0 ; i < USE_NUM_ENTRIES ; i ++ )
    {
        tower [ i ] . true_address = & smmu [
                    tower [ i ] . index ] ;
        assert (( tower [ i ] . index + tower [ i ] . size ) < max_index ) ;
        assert ( tower [ i ] . index >= min_index ) ;

#ifdef VERBOSE
        printf ( "Block: " ) ;
        printhex ( i ) ;
        printf ( " address " ) ;
        printhex ( (unsigned) tower [ i ] . true_address ) ;
        printf ( " begin " ) ;
        printhex ( tower [ i ] . index ) ;
        printf ( " end " ) ;
        printhex ( tower [ i ] . index + tower [ i ] . size ) ;
        printf ( " copy_loc " ) ;
        printhex ( tower [ i ] . copy_loc ) ;
        printf ( "\r\n" ) ;
        assert ( tower [ i ] . size != 0 ) ;
#endif
    }
    for ( i = 0 ; i < max_index ; i ++ )
    {
        shadow [ i ] = i ^ KEY2 ^ KEY1 ;
        smmu [ i ] = i ^ KEY2 ;
    }
    space += max_index ;
    assert ( space <= SP_SIZE ) ;
    loaded = 0 ;

    for ( i = 0 ; i < NUM_CYCLES ; )
    {
        unsigned    rngv = rng () ;
        unsigned    warm_up = rngv & 1 ;

        printf ( "\rMajor cycle " ) ;
        printhex ( i ) ;
        printf ( " " ) ;

        if (( rngv >> 20 ) & 1 )
        {
#if XPAR_MICROBLAZE_0_USE_ICACHE
            microblaze_disable_icache();
            microblaze_enable_icache();
#endif
        }
        for ( j = 0 ; j < USE_NUM_ENTRIES ; j ++ )
        {
            assert ( ! tower [ j ] . loaded ) ;
        }
        if ( rngv & 2 )
        {
            if ( loaded )
            {
                HW_CLOSE ( NUM_ENTRIES - 1 ) ;
                loaded = 0 ;
                assert ( num_rw_open > 0 ) ;
                num_rw_open -- ;
            } else {
                j = HW_OPEN ( shadow , 
                        max_index * sizeof ( data_t ) , copy_loc ) ;
                assert ( j == ( NUM_ENTRIES - 1 ) ) ;
                loaded = 1 ;
                assert ( num_rw_open < NUM_RW_ENTRIES ) ;
                num_rw_open ++ ;
            }
        }

        for ( sc = 0 ; sc < ( NUM_SUB_CYCLES + USE_NUM_ENTRIES ) ; sc ++, i ++ )
        {
            unsigned    apply_to , index ;
            unsigned    error_count , k , rw_mode ;

            rngv = rng () ;
#ifdef VERBOSE
            printf ( "Start minor cycle " ) ;
            printhex ( i ) ;
            printf ( " " ) ;
            printhex ( rngv ) ;
            printf ( "\r\n" ) ;
#endif
            rw_mode = rngv % 2 ;
            rngv /= 2 ;

            if ( sc >= NUM_SUB_CYCLES )
            {
                apply_to = sc - NUM_SUB_CYCLES ;
                assert ( apply_to < USE_NUM_ENTRIES ) ;
            } else if (( sc < USE_NUM_ENTRIES )
            && ( warm_up ))
            {
                apply_to = i % USE_NUM_ENTRIES ;
            } else {
                apply_to = rngv % USE_NUM_ENTRIES ;
            }

            while ( rngv > ( USE_NUM_ENTRIES * 8 ))
            {
                rngv /= USE_NUM_ENTRIES ;
                switch ( rngv % 8 )
                {
                    case 0 :    index = tower [ apply_to ] . index - 1 ;
                                break ;
                    case 1 :    index = tower [ apply_to ] . index ;
                                break ;
                    case 2 :    index = tower [ apply_to ] . index +
                                        tower [ apply_to ] . size ;
                                break ;
                    case 3 :    index = tower [ apply_to ] . index +
                                        tower [ apply_to ] . size - 1 ;
                                break ;
                    default :   index = ( rngv / 8 ) % max_index ;
                                rngv /= max_index ;
                                break ;
                }
                rngv /= 8 ;
                assert ( index < max_index ) ;
                k = KEY1 ;
                switch ( width )
                {
                    case 4 :    shadow [ index ] = ( i + 1 ) ^ k ;
                                smmu [ index ] = i + 1 ;
#ifdef VERBOSE
                                printf ( "store word at " ) ;
                                printhex ( (unsigned) & smmu [ index ] ) ;
                                printf ( " value " ) ;
                                printhex ( i + 1 ) ;
                                printf ( "\r\n" ) ;
#endif
                                break ;
                    case 2 :    index *= 2 ;
                                index += rngv % 2 ;
                                k = k >> (( 1 - ( rngv % 2 )) * 16 ) ;
                                rngv /= 2 ;
                                shadow2 [ index ] = ( i + 1 ) ^ k ;
                                smmu2 [ index ] = i + 1 ;
                                break ;
                    case 1 :    index *= 4 ;
                                index += rngv % 4 ;
                                k = k >> (( 3 - ( rngv % 4 )) * 8 ) ;
                                rngv /= 4 ;
                                shadow4 [ index ] = ( i + 1 ) ^ k ;
                                smmu4 [ index ] = i + 1 ;
                                break ;
                    default :   assert ( 0 ) ;
                                break ;
                }
            }

            if ( ! tower [ apply_to ] . loaded )
            {
                if ( sc < NUM_SUB_CYCLES )
                {
                    unsigned true_size ;

                    true_size = tower [ apply_to ] . size * sizeof ( data_t ) ;
                    if ( num_ro_open == NUM_RO_ENTRIES )
                    {
                        /* Must be RW */
                        rw_mode = 1 ;
                    } else if ( num_rw_open == NUM_RW_ENTRIES )
                    {
                        /* Must be RO */
                        rw_mode = 0 ;
                    }
#ifdef VERBOSE
                    printf ( " open block " ) ;
                    printhex ( apply_to ) ;
                    printf ( " start " ) ;
                    printhex ( (unsigned) tower [ apply_to ] . true_address ) ;
                    printf ( " size " ) ;
                    printhex ( true_size ) ;
                    printf ( " copy_loc " ) ;
                    printhex ( (unsigned) tower [ apply_to ] . copy_loc ) ;
                    printf ( "%s", rw_mode ? " RW" : " RO" ) ;
#endif
                    if ( rw_mode )
                    {
                        /* Read-write open */
                        assert ( num_rw_open < NUM_RW_ENTRIES ) ;
                        num_rw_open ++ ;
                        tower [ apply_to ] . ref =
                            HW_OPEN ( tower [ apply_to ] . true_address ,
                                    true_size ,
                                    tower [ apply_to ] . copy_loc ) ;
                        assert ( tower [ apply_to ] . ref >= FIRST_RW_ENTRY ) ;
                    } else {
                        /* Read-only open */
                        assert ( num_ro_open < NUM_RO_ENTRIES ) ;
                        num_ro_open ++ ;
                        tower [ apply_to ] . ref =
                            HW_OPEN_RO ( tower [ apply_to ] . true_address ,
                                    true_size ,
                                    tower [ apply_to ] . copy_loc ) ;
                        assert ( tower [ apply_to ] . ref < FIRST_RW_ENTRY ) ;
                        assert ( tower [ apply_to ] . ref != 0 ) ;
                    }

                    tower [ apply_to ] . loaded = 1 ;

#ifdef VERBOSE
                    printf ( " ref " ) ;
                    printhex ( tower [ apply_to ] . ref ) ;
                    printf ( "\r\n" ) ;
#endif
                    assert ( tower [ apply_to ] . ref < NUM_ENTRIES ) ;
                }
            } else {
                /* Unload it */
                HW_CLOSE ( tower [ apply_to ] . ref ) ;
                tower [ apply_to ] . loaded = 0 ;
                if ( tower [ apply_to ] . ref < FIRST_RW_ENTRY )
                {
                    /* Read only */
                    assert ( num_ro_open > 0 ) ;
                    num_ro_open -- ;
                } else {
                    assert ( num_rw_open > 0 ) ;
                    num_rw_open -- ;
                }
#ifdef VERBOSE
                printf ( " close block " ) ;
                printhex ( apply_to ) ;
                printf ( " ref " ) ;
                printhex ( tower [ apply_to ] . ref ) ;
                printf ( "\r\n" ) ;
#endif
            }

            error_count = 0 ;

#ifndef VERBOSE
            if (( i % 0x400 ) == 0 )
            {
                printhex ( i ) ;
                printf ( "\x08\x08\x08\x08\x08\x08\x08\x08" ) ;
            }
#endif
            for ( j = min_index ; j < max_index ; j ++ )
            {
                unsigned a = shadow [ j ] ;
                unsigned b = smmu [ j ] ;

                if (( a ^ b ) != KEY1 )
                {
                    error_count ++ ;
                    if ( error_count < MAX_ERRORS )
                    {
                        printf ( "Error " ) ;
                        printhex ( error_count ) ;
                        printf ( " detected at index " ) ;
                        printhex ( j ) ;
                        printf ( " address " ) ;
                        printhex ( (unsigned) & smmu [ j ] ) ;
                        printf ( "\r\n  shadow is " ) ;
                        printhex ( a ^ KEY1 ) ;
                        printf ( " smmu is " ) ;
                        printhex ( b ) ;
                        printf ( "\r\n" ) ;
                    } else if ( error_count == MAX_ERRORS )
                    {
                        printf ( "more errors...\r\n" ) ;
                    }
                }
            }
            if ( error_count != 0 )
            {
                printf ( "\r\nTowering test: cycle " ) ;
                printhex ( i ) ;
                printf ( " - " ) ;
                printhex ( error_count ) ;
                printf ( " errors detected." ) ;
                printf ( "\r\n" ) ;
                return 0 ;
            }
        }
    }
    if ( loaded )
    {
        HW_CLOSE ( NUM_ENTRIES - 1 ) ;
        loaded = 0 ;
        assert ( num_rw_open == 1 ) ;
        num_rw_open -- ;
    }
    assert ( num_rw_open == 0 ) ;
    assert ( num_ro_open == 0 ) ;
    return 1 ;
}

static int Run_Test ( char * name , Size_Record * tower )
{
    return Run_Test_W ( name , tower , 4 ) 
        && Run_Test_W ( name , tower , 1 ) 
        && Run_Test_W ( name , tower , 2 ) ;
}

